# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json
import os
import time

from typing import Optional
from pathlib import Path

from hvcc.core.hv2ir.HeavyException import HeavyException
from hvcc.core.hv2ir.HeavyParser import HeavyParser

from hvcc.types.compiler import CompilerResp, CompilerNotif, CompilerMsg


class hv2ir:

    @classmethod
    def compile(
        cls,
        hv_file: Path,
        ir_file: Path,
        patch_name: Optional[str] = None,
        verbose: bool = False
    ) -> CompilerResp:
        """ Compiles a HeavyLang file into a HeavyIR file.
            Returns a tuple of compile time in seconds, a notification dictionary,
            and a heavy object counter.
        """

        # keep track of the total compile time
        tick = time.time()

        hv_file = hv_file.expanduser().absolute()
        ir_file = ir_file.expanduser().absolute()

        try:
            # parse heavy file
            hv_graph = HeavyParser.graph_from_file(hv_file=hv_file, xname=patch_name)
        except HeavyException as e:
            return CompilerResp(
                stage="hv2ir",
                compile_time=time.time() - tick,
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    errors=[CompilerMsg(message=e.message)],
                    warnings=[]
                ),
                in_file=hv_file,
                in_dir=hv_file.parent,
                out_file=ir_file,
                out_dir=ir_file.parent
            )

        try:
            # get a counter of all heavy objects
            hv_counter = hv_graph.get_object_counter(recursive=True)

            # prepare the graph for exporting
            hv_graph.prepare()

            # ensure that the output directory exists
            if not ir_file.parent.exists():
                os.makedirs(ir_file.parent)

            # generate Heavy.IR
            ir = hv_graph.to_ir()
        except HeavyException as e:
            return CompilerResp(
                stage="hv2ir",
                compile_time=time.time() - tick,
                notifs=CompilerNotif(
                    has_error=True,
                    exception=e,
                    errors=[CompilerMsg(message=e.message)],
                    warnings=[]
                ),
                in_file=hv_file,
                in_dir=hv_file.parent,
                out_file=ir_file,
                out_dir=ir_file.parent,
                obj_counter=hv_counter
            )

        if ir is not None:
            # write the hv.ir file
            with open(ir_file, "w") as f:
                json.dump(ir.model_dump(), f, indent=4)

            if verbose and ir is not None:
                if len(ir.signal.processOrder) > 0:
                    print("")
                    print("=== Signal Order ===")
                    for so in ir.signal.processOrder:
                        o = ir.objects[so.id]
                        if len(o.args) > 0:
                            print("{0} {{{1}}}".format(
                                o.type,
                                " ".join([f"{k}:{v}" for k, v in o.args.items()])))
                        else:
                            print(o.type)

        return CompilerResp(
            stage="hv2ir",
            compile_time=time.time() - tick,  # record the total compile time
            notifs=hv_graph.get_notices(),
            in_file=hv_file,
            in_dir=hv_file.parent,
            out_file=ir_file,
            out_dir=ir_file.parent,
            obj_counter=hv_counter,
            ir=ir
        )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A C-language compiler for the Heavy audio programming language.")
    parser.add_argument(
        "hv_path",
        help="The patch to the top-level patch to compile.")
    parser.add_argument(
        "--hv_ir_path",
        default="./heavy.hv.ir.json",
        help="The output path of the hv.ir.json file.")
    parser.add_argument(
        "--name",
        default="heavy",
        help="")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    d = hv2ir.compile(
        hv_file=args.hv_path,
        ir_file=args.hv_ir_path,
        patch_name=args.name,
        verbose=args.verbose)

    if args.verbose:
        print(f"Total hv2ir time: {(d.compile_time * 1000):.2f}ms")


if __name__ == "__main__":
    main()



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from .PdObject import PdObject

from typing import Optional

from hvcc.types.Heavy import Heavy, HvPos


class PdExprObject(PdObject):
    """
    Limitations (compared to vanilla pd):
    - only supports a single expression, but we can split into multiple objects in PdParser
    - Available pd docs/examples say expr support up to 100 variables
      This version currently supports up to 100 variables as defined in HvControlExpr.h
    - I don't know what pd-expr does with strings, haven't experimented
      and haven't given it any thought yet here
    """

    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[list] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        """
            Validate the expr object and any heavy restrictions, then
            convert it directly into a HeavyIR object.
        """
        assert obj_type in ["expr", "expr~"]
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        # turn the arguments into a list of expressions, but only one for now
        if len(self.obj_args) == 0:
            self.add_error("Empty expression")

        expressions = [e.strip() for e in " ".join(self.obj_args).split("\\;")]
        if len(expressions) > 1:
            self.add_error("Heavy expr does not support multiple expressions")
        else:
            self.expressions = expressions

        # this is overridden by PdParser
        self.num_inlets = 1

    def validate_configuration(self) -> None:
        # things that could be validated:
        # - inlet count/types match variables in the expression(s)
        pass

    def to_hv(self) -> Heavy:
        return Heavy(
            type=f"__{self.obj_type}",
            args={
                "expressions": self.expressions,
                "num_inlets": self.num_inlets
            },
            properties=HvPos(x=self.pos_x, y=self.pos_y)
        )



# Copyright (C) 2022-2025 Daniel Billotte, Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from typing import Callable, Dict, List, Tuple

from .HeavyObject import HeavyObject


class ControlExpr(HeavyObject):
    c_struct = "ControlExpr"
    preamble = "cExpr"

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvControlExpr.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {"HvControlExpr.h", "HvControlExpr.c"}

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: Dict) -> List[str]:
        """ (Per object) code that gets inserted into from the Heavy_heavy ctor.
            Only if "ir[init]" == true
        """

        eval_f = f"&Heavy_{{{{name}}}}::{cls.preamble}_{obj_id}_evaluate"
        return [f"cExpr_init(&cExpr_{obj_id}, {eval_f});"]

    @classmethod
    def get_C_def(cls, obj_type: str, obj_id: str) -> List[str]:
        """ (Per object) code that gets inserted into the header file
            Only if "ir[init]" == true
        """

        lines = super().get_C_def(obj_type, obj_id)
        lines.append(f"static float {cls.preamble}_{obj_id}_evaluate(const float* args);")
        return lines

    @classmethod
    def get_C_onMessage(cls, obj_type: str, obj_id: str, inlet_index: int, args: Dict) -> List[str]:
        """ (Per object) code that gets inserted into the c<PREAMBLE>_<OBJID>_sendMessage
            method in the .cpp file

            The get_C_onMessage method returns the code that will get inserted into
            the cReceive_<UID>_sendMessage method
        """

        return [
            "cExpr_onMessage(_c, &Context(_c)->cExpr_{0}, {1}, m, &cExpr_{0}_sendMessage);".format(
                obj_id,
                inlet_index)
        ]

    @classmethod
    def get_C_impl(
        cls,
        obj_type: str,
        obj_id: str,
        on_message_list: List,
        get_obj_class: Callable,
        objects: Dict,
        args: Dict
    ) -> List[str]:
        """ (Per object) this creates the _sendMessage function that other objects use to
            send messages to this object.
        """

        lines = super().get_C_impl(obj_type, obj_id, on_message_list, get_obj_class, objects, args)
        expr = args["expressions"][0]
        bound_expr = bind_expr(expr, "args")
        lines.extend([
            "",
            f"float Heavy_{{{{name}}}}::{cls.preamble}_{obj_id}_evaluate(const float* args) {{",
            f"\treturn {bound_expr};",
            "}",
        ])
        return lines


def var_n(a_name: str, var: str) -> str:
    """ Rewrite the input expression into one that uses local variables
        that have been cast to either float or int

        todo(dgb): need to handle the 's' type
    """
    parts = re.match(r"\$([fi])(\d+)", var)
    assert parts
    type = "float" if parts[1] == "f" else "int"
    return f"(({type})({a_name}[{int(parts[2])-1}]))"


def internal_expr(exp: str) -> str:
    """ Convert function names to C or internal names
    """
    internal: List[Tuple[str, str]] = [
        (r"\\,",           ","),
        (r"\bfact(?=\()",  "expr_fact"),
        (r"\bimodf(?=\()", "expr_imodf"),
    ]

    for r in internal + hv_utils:
        exp = re.sub(r[0], r[1], exp)

    return exp


def bind_expr(exp: str = "$f1+2", a_name: str = "a") -> str:
    vars = re.findall(r"\$[fis]\d+", exp)
    exp = internal_expr(exp)

    if vars:
        # reverse list so we start replacing the bigger variables
        # and don't get conflicts. eg. $f11 vs $f1
        vars.sort(reverse=True)
        for var in vars:
            exp = exp.replace(var, var_n(a_name, var))

    return exp


hv_utils: List[Tuple[str, str]] = [
    (r"\babs(?=\()",        "hv_abs_f"),
    (r"\bacos(?=\()",       "hv_acos_f"),
    (r"\bacosh(?=\()",      "hv_acosh_f"),
    (r"\basin(?=\()",       "hv_asin_f"),
    (r"\basinh(?=\()",      "hv_asinh_f"),
    (r"\batan(?=\()",       "hv_atan_f"),
    (r"\batanh(?=\()",      "hv_atanh_f"),
    (r"\batan2(?=\()",      "hv_atan2_f"),
    (r"\bcbrt(?=\()",       "hv_cbrt_f"),
    (r"\bceil(?=\()",       "hv_ceil_f"),
    (r"\bcopysign(?=\()",   "hv_copysign_f"),
    (r"\bcos(?=\()",        "hv_cos_f"),
    (r"\bcosh(?=\()",       "hv_cosh_f"),
    (r"\berf(?=\()",        "hv_erf_f"),
    (r"\berfc(?=\()",       "hv_erfc_f"),
    (r"\bexp(?=\()",        "hv_exp_f"),
    (r"\bexpm1(?=\()",      "hv_expm1_f"),
    (r"\bfinite(?=\()",     "hv_finite_f"),
    (r"\bfloor(?=\()",      "hv_floor_f"),
    (r"\bif(?=\()",         "hv_if_f"),
    (r"\bisinf(?=\()",      "hv_isinf_f"),
    (r"\bisnan(?=\()",      "hv_isnan_f"),
    (r"\bfmod(?=\()",       "hv_fmod_f"),
    (r"\bldexp(?=\()",      "hv_ldexp_f"),
    (r"\bln(?=\()",         "hv_ln_f"),
    (r"\blog(?=\()",        "hv_log_f"),
    (r"\blog10(?=\()",      "hv_log10_f"),
    (r"\blog1p(?=\()",      "hv_log1p_f"),
    (r"\bmax(?=\()",        "hv_max_f"),
    (r"\bmin(?=\()",        "hv_min_f"),
    (r"\bmodf(?=\()",       "hv_modf_f"),
    (r"\bpow(?=\()",        "hv_pow_f"),
    (r"\bremainder(?=\()",  "hv_remainder_f"),
    (r"\brint(?=\()",       "hv_rint_f"),
    (r"\bround(?=\()",      "hv_round_f"),
    (r"\bsin(?=\()",        "hv_sin_f"),
    (r"\bsinh(?=\()",       "hv_sinh_f"),
    (r"\bsqrt(?=\()",       "hv_sqrt_f"),
    (r"\btan(?=\()",        "hv_tan_f"),
    (r"\btanh(?=\()",       "hv_tanh_f"),
]



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from typing import Optional
from pathlib import Path

from .PdObject import PdObject
from .HeavyObject import HeavyObject

from hvcc.types.Heavy import Heavy


class HeavyGraph(PdObject):
    def __init__(
        self,
        hv_path: Path,
        obj_args: Optional[list] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        super().__init__(hv_path.name.split(".")[0], obj_args, pos_x, pos_y)

        # read the heavy graph
        with open(hv_path, "r") as f:
            self.hv_json = Heavy(**json.load(f))

        # parse the heavy data structure to determine the outlet connection type
        outlets = [o for o in self.hv_json.objects.values() if o.type == "outlet"]
        outlets = sorted(outlets, key=lambda o: o.args["index"] if isinstance(o.args, dict) else o.args[0])
        self.__outlet_connection_types: list[str] = [o.args["type"] for o in outlets if isinstance(o.args, dict)]

        # resolve the arguments
        for i, a in enumerate(self.hv_json.args):
            if i < len(self.obj_args):
                arg_value = self.obj_args[i]
            elif a["required"]:
                self.add_error(f"Required argument \"{a['name']}\" not found.")
                continue
            else:
                arg_value = a["default"]

            try:
                arg_value = HeavyObject.force_arg_type(arg_value, a["value_type"])
            except Exception as e:
                self.add_error(
                    f"Heavy {self.obj_type} cannot convert argument \"{a['name']}\""
                    f" with value \"{arg_value}\" to type {a['value_type']}: {e}")

            # resolve all arguments for each object in the graph
            for o in self.hv_json.objects.values():
                assert isinstance(o.args, dict)
                for k, v in o.args.items():
                    # TODO(mhroth): make resolution more robust
                    if v == "$" + a["name"]:
                        o.args[k] = arg_value

        # reset all arguments, as they have all been resolved
        # any required arguments would break hv2ir as they will no longer
        # be supplied (because they are resolved)
        self.hv_json.args = []

    def get_outlet_connection_type(self, outlet_index: int) -> str:
        return self.__outlet_connection_types[outlet_index]

    def to_hv(self) -> Heavy:
        return self.hv_json



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


# moved to HeavyParser.py because of circular dependency

# from pathlib import Path

# from .HeavyException import HeavyException
# from .HeavyIrObject import HeavyIrObject
# from .HeavyLangObject import HeavyLangObject

# class HLangIf(HeavyLangObject):
#     """ Translates HeavyLang object [if] to HeavyIR [if] or [if~].
#     """

#     def __init__(
#         self,
#         obj_type: str,
#         args: Dict,
#         graph: HeavyGraph,
#         annotations: Optional[Dict] = None
#     ) -> None:
#         assert obj_type == "if"
#         super().__init__("if", args, graph, num_inlets=2, num_outlets=2, annotations=annotations)

#     def reduce(self) -> Tuple[Set, List]:
#         if self.has_inlet_connection_format(["cc", "_c", "c_", "__"]):
#             x = HeavyIrObject("__if", self.args)
#         elif self.has_inlet_connection_format("ff"):
#             # TODO(mhroth): implement this
#             x = HeavyParser.graph_from_file(Path("./hvlib/if~f.hv.json"))
#         elif self.has_inlet_connection_format("ii"):
#             # TODO(mhroth): implement this
#             x = HeavyParser.graph_from_file(Path("./hvlib/if~i.hv.json"))
#         else:
#             fmt = self._get_connection_format(self.inlet_connections)
#             raise HeavyException(f"Unhandled connection configuration to object [if]: {fmt}")

#         return ({x}, self.get_connection_move_list(x))



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import argparse
import json

from collections import Counter, defaultdict
from typing import Dict
from pathlib import Path

from hvcc.types.IR import HeavyIRType, IRGraph


class ir2c_perf:

    @classmethod
    def perf(
        cls,
        ir: IRGraph,
        blocksize: int = 512,
        mhz: int = 1000,
        verbose: bool = False
    ) -> Dict[str, Dict[str, float]]:
        # read the hv.ir.json file
        with open(Path(Path(__file__).parent, "../../core/json/heavy.ir.json"), "r") as f:
            HEAVY_IR_JSON = HeavyIRType(**json.load(f)).root

        objects: Counter = Counter()
        perf: Dict[str, float] = defaultdict(float)
        per_object_perf: Dict[str, Dict[str, float]] = defaultdict(lambda: defaultdict(float))

        for o in ir.signal.processOrder:
            obj_id = o.id
            obj_type = ir.objects[obj_id].type
            if obj_type in HEAVY_IR_JSON.keys():
                objects[obj_type] += 1
                obj_perf = HEAVY_IR_JSON[obj_type].perf
                assert obj_perf is not None
                c = defaultdict(float, **obj_perf.model_dump())
                for k, v in c.items():
                    perf[k] = perf[k] + v
                    per_object_perf[obj_type][k] = per_object_perf[obj_type][k] + v
            else:
                print(f"ERROR: Unknown object type {obj_type}")

        if verbose:
            print("AVX: {0} cycles / {1} cycles per frame".format(perf["avx"], perf["avx"] / 8.0))
            print("     {0} frames @ {1}MHz >= {2:.2f}us".format(
                blocksize,
                mhz,
                blocksize * perf["avx"] / 8.0 / mhz))

            print()  # new line

            print("SSE: {0} cycles / {1} cycles per frame".format(perf["sse"], perf["sse"] / 4.0))
            print("     {0} frames @ {1}MHz >= {2:.2f}us".format(
                blocksize,
                mhz,
                blocksize * perf["sse"] / 4.0 / mhz))

            print()  # new line

            print("NEON: {0} cycles / {1} cycles per frame".format(perf["neon"], perf["neon"] / 4.0))
            print("     {0} frames @ {1}MHz >= {2:.2f}us".format(
                blocksize,
                mhz,
                blocksize * perf["neon"] / 4.0 / mhz))

            print()  # new line

            print("{0:<4} {1:<5} {2:<16} {3}".format("CPU%", "#Objs", "Object Type", "Performance"))
            print("==== ===== ================ ===========")

            # print object in order of highest load
            items = per_object_perf.items()
            # items.sort(key=lambda o: o[1]["avx"], reverse=True)
            for k, v in items:
                if perf["avx"] > 0:
                    print(
                        "{2:>2.2g}%  {3:<5} {0:<16} {1}".format(
                            k, dict(v), int(100.0 * v["avx"] / perf["avx"]), objects[k])
                    )

        return per_object_perf


def main() -> None:
    parser = argparse.ArgumentParser(
        description="A Heavy.IR to C-language translator.")
    parser.add_argument(
        "hv_ir_path",
        help="The path to the Heavy.IR file to read.")
    parser.add_argument("--mhz", default=1000, type=float, help="the CPU clock frequency in MHz")
    parser.add_argument("--blocksize", default=64, type=int, help="the number of frames per block")
    parser.add_argument("-v", "--verbose", action="count")
    args = parser.parse_args()

    # read the hv.ir.json file
    with open(args.hv_ir_path, "r") as f:
        ir = json.load(f)

    ir2c_perf.perf(ir, args.blocksize, args.mhz, args.verbose)


if __name__ == "__main__":
    main()



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional

from .PdObject import PdObject

from hvcc.types.Heavy import Heavy, HvPos


class PdLetObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[list] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type in {"inlet", "inlet~", "outlet", "outlet~"}
        super().__init__(obj_type, obj_args, pos_x, pos_y)
        self.let_index = 0

    def get_outlet_connection_type(self, outlet_index: int) -> Optional[str]:
        if len(self.obj_args) > 0 and self.obj_args[0] in {"-->", "~f>", "~i>", "-~>"}:
            return self.obj_args[0]
        else:
            return super().get_outlet_connection_type(outlet_index)

    def to_hv(self) -> Heavy:
        return Heavy(
            type=self.obj_type.strip("~"),
            args={
                "name": "",  # Pd does not give an inlet name
                "index": self.let_index,
                "type": self.get_outlet_connection_type(self.let_index)
            },
            properties=HvPos(x=self.pos_x, y=self.pos_y)
        )



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import Counter
from typing import Optional

from .NotificationEnum import NotificationEnum
from .PdObject import PdObject

from hvcc.types.Heavy import Heavy, HvPos, HvConn, HvConnFrom, HvConnTo


class PdSelectObject(PdObject):
    def __init__(
        self,
        obj_type: str,
        obj_args: Optional[list] = None,
        pos_x: int = 0,
        pos_y: int = 0
    ) -> None:
        assert obj_type in {"select", "sel"}
        super().__init__(obj_type, obj_args, pos_x, pos_y)

        if not obj_args:
            obj_args = []

        if len(self.obj_args) == 0:
            self.add_error("At least one argument is required.")
        if len(set(self.obj_args)) != len(obj_args):
            c = Counter(obj_args).most_common(1)
            self.add_error(
                f"All arguments to [select] must be unique. Argument \"{c[0][0]}\" is repeated {c[0][1]} times.",
                NotificationEnum.ERROR_UNIQUE_ARGUMENTS_REQUIRED)

        # convert to obj_args to mixedarray, such that correct switchcase hash
        # is generated
        for i, a in enumerate(self.obj_args):
            try:
                self.obj_args[i] = float(a)
            except Exception:
                pass

    def validate_configuration(self) -> None:
        if len(self._inlet_connections.get("1", [])) > 0:
            self.add_warning("The right inlet of select is not supported. It will not do anything.")

    def to_hv(self) -> Heavy:
        """ Creates a graph dynamically based on the number of arguments.
            An unconnected right inlet is added.

            [inlet]                                         [inlet]
            |
            [@hv_obj switchcase [arg list (N elements)]           ]
            |                             |                       |
            [@hv_obj __cast_b]            [@hv_obj __cast_b]      |
            |                             |                       |
            [outlet_0]                    [outlet_N-1]            [outlet_right]
        """

        route_graph = Heavy(
            type="graph",
            args=[],
            objects={
                "inlet": Heavy(
                    type="inlet",
                    args={
                        "type": "-->",
                        "index": 0
                    }
                ),
                "inlet_right": Heavy(
                    type="inlet",
                    args={
                        "type": "-->",
                        "index": 1
                    }
                ),
                "switchcase": Heavy(
                    type="__switchcase",
                    args={
                        "cases": self.obj_args
                    }
                ),
                "outlet_right": Heavy(
                    type="outlet",
                    args={
                        "type": "-->",
                        "index": len(self.obj_args)
                    }
                ),
            },
            connections=[
                HvConn(
                    type="-->",
                    conn_from=HvConnFrom(id="inlet", outlet=0),
                    conn_to=HvConnTo(id="switchcase", inlet=0)
                ),
                HvConn(
                    type="-->",
                    conn_from=HvConnFrom(id="switchcase", outlet=len(self.obj_args)),
                    conn_to=HvConnTo(id="outlet_right", inlet=0)
                )
            ],
            properties=HvPos(x=self.pos_x, y=self.pos_y)
        )

        for i in range(len(self.obj_args)):
            # add __cast_b to graph
            route_graph.objects[f"__cast_b_{i}"] = Heavy(
                type="__cast_b"
            )

            # add outlets to graph
            route_graph.objects[f"outlet_{i}"] = Heavy(
                type="outlet",
                args={
                    "type": "-->",
                    "index": i
                }
            )

            # add connection from switchcase to slice
            route_graph.connections.append(
                HvConn(
                    type="-->",
                    conn_from=HvConnFrom(id="switchcase", outlet=i),
                    conn_to=HvConnTo(id=f"__cast_b_{i}", inlet=0)
                )
            )

            # add connection from slice to outlet
            route_graph.connections.append(
                HvConn(
                    type="-->",
                    conn_from=HvConnFrom(id=f"__cast_b_{i}", outlet=0),
                    conn_to=HvConnTo(id=f"outlet_{i}", inlet=0)
                )
            )

        return route_graph



# MIT License
#
# Copyright (c) 2026 Jaffx Audio Inc.
# Copyright (c) 2026 Wasted Audio
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


import re
import json

from enum import Enum
from pathlib import Path


""" Vendored and modified from MicroNAM https://github.com/jaffco/MicroNAM
"""


class ModelNet(Enum):
    Nano = 1
    Feather = 2
    Lite = 3
    Standard = 4


def to_symbol_stem(nam_path: Path) -> str:
    stem = nam_path.stem.strip()
    if not stem:
        return "NamModel"

    # Split on non-alphanumeric boundaries, then form PascalCase.
    raw_parts = [p for p in re.split(r"[^A-Za-z0-9]+", stem) if p]
    if not raw_parts:
        return "NamModel"

    parts = []
    for part in raw_parts:
        if part[0].isdigit():
            part = f"M{part}"
        parts.append(part[0].upper() + part[1:])

    return "".join(parts)


def format_float(value: float) -> str:
    # Keep enough precision for model weights while staying compact.
    return f"{format(value, '.17g')}f"


def load_nam_file(nam_file: Path) -> tuple[ModelNet, int, list[float]]:
    with nam_file.open("r", encoding="utf-8") as f:
        payload = json.load(f)

    if "weights" not in payload.keys() or not isinstance(payload["weights"], list):
        raise ValueError("Input .nam file must contain a 'weights' array")

    weights = payload["weights"]
    for idx, value in enumerate(weights):
        if not isinstance(value, (int, float)):
            raise ValueError(f"weights[{idx}] is not numeric")

    weights_len = len(weights)

    if weights_len == 842:
        model_net = ModelNet.Nano
    elif weights_len == 3026:
        model_net = ModelNet.Feather
    elif weights_len == 6554:
        model_net = ModelNet.Lite
    elif weights_len == 13802:
        model_net = ModelNet.Standard
    else:
        raise ValueError(f"Invalid weights length: {weights_len}")

    return model_net, weights_len, weights


def convert_nam_to_header(nam_file: Path) -> str:
    symbol_stem = to_symbol_stem(nam_file)
    count_name = f"{symbol_stem}WeightsCount"
    weights_name = f"{symbol_stem}Weights"

    _, weights_len, weights = load_nam_file(nam_file)

    lines = [
        "#pragma once",
        "",
        f"static const unsigned int {count_name} = {weights_len};",
        f"static const float {weights_name}[{count_name}] = {{",
    ]

    row = []
    for idx, value in enumerate(weights, start=1):
        row.append(format_float(float(value)))
        if idx % 6 == 0:
            lines.append("    " + ", ".join(row) + ",")
            row = []

    if row:
        lines.append("    " + ", ".join(row) + ",")

    lines.append("};")
    lines.append("")

    return "\n".join(lines)



# Heavy Compiler Collection
# Copyright (C) 2025-2026 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

import re

from collections import Counter
from typing import Generator, Optional, Union
from pathlib import Path

from hvcc.interpreters.pd2hv.PdParser import PdParser
from hvcc.types.GUI import (
    Size, Coords, Font, LabelShow, LabelPos, Label, Color, Canvas,
    Bang, Toggle, VRadio, HRadio, VSlider, HSlider, Knob,  Number, Float,
    Comment, GUIObjects, Graph, GraphRoot
)


class PdGUIParser(PdParser):
    # retain width by overriding this regex with a pattern that never matches
    RE_WIDTH = re.compile(r"$^")

    def __init__(self) -> None:
        # the current global value of $0
        # Note(joe): set a high starting value to avoid potential user naming conflicts
        self.__DOLLAR_ZERO = 1000
        self.object_counter: Counter = Counter()

        # search paths at this graph level
        self.search_paths: list[Path] = []

    def gui_from_file(
        self,
        file_path: Path,
        obj_args: Optional[list] = None,
        is_root: bool = True
    ) -> tuple[Union[Graph, GraphRoot], bool]:
        if is_root:
            self.search_paths.append(file_path)

        file_iterator = self.get_pd_line(file_path)
        canvas_line: str = file_iterator.__next__()

        self.__DOLLAR_ZERO += 1  # increment $0
        graph_args = [self.__DOLLAR_ZERO] + (obj_args or [])

        if not canvas_line.startswith("#N canvas"):
            raise Exception(f"Pd files must begin with \"#N canvas\": {canvas_line}")

        g, gop = self.gui_from_canvas(
            file_iterator,
            canvas_line,
            graph_args,
            file_path,
            is_root
        )

        return g, gop

    def gui_from_canvas(
        self,
        file_iterator: Generator,
        canvas_line: str,
        graph_args: list,
        pd_path: Path,
        is_root: bool = False
    ) -> tuple[Union[Graph, GraphRoot], bool]:

        objects: list[GUIObjects] = []
        graphs: list[Graph] = []
        x: Optional[GUIObjects] = None

        # graph on parent settings
        gop: bool = False
        gop_start: Coords = Coords()
        gop_size: Size = Size()
        obj_type = ""
        obj_args = []

        try:
            for li in file_iterator:
                line = self.split_line(li)

                if line[0] == "#N":
                    if line[1] == "canvas":
                        g, gop = self.gui_from_canvas(
                            file_iterator=file_iterator,
                            canvas_line=li,
                            graph_args=graph_args,
                            pd_path=pd_path
                        )
                        if gop:
                            assert isinstance(g, Graph)
                            graphs.append(g)

                elif line[0] == "#X":
                    if line[1] == "coords":
                        try:
                            if int(line[8]) > 0:
                                # canvas is active
                                gop = True
                                gop_start = Coords(x=int(line[9]), y=int(line[10]))
                                gop_size = Size(x=int(line[6]), y=int(line[7]))
                        except IndexError:
                            continue

                    elif line[1] == "restore" and gop:
                        objects = self.filter_invisible_objects(objects, gop_start, gop_size)
                        graphs = self.filter_invisible_graphs(graphs, gop_start, gop_size)
                        self.object_counter["graph"] += 1

                        return Graph(
                            id=f"graph{self.object_counter['graph']}",
                            position=Coords(
                                x=int(line[2]),
                                y=int(line[3])
                            ),
                            gop_start=gop_start,
                            gop_size=gop_size,
                            objects=objects,
                            graphs=graphs
                        ), gop

                    elif line[1] == "text":
                        x = self.add_comment(line)

                    elif line[1] == "floatatom":
                        obj_args = self.__resolve_object_args(
                            obj_args=line[1:],
                            graph_args=graph_args,
                            is_root=is_root
                        )

                        # replace args with resolved args
                        line = line[:1] + obj_args
                        x = self.add_float(line)

                    elif line[1] == "obj":
                        if len(line) > 4:
                            obj_type = line[4]
                            # sometimes objects have $ arguments in them as well
                            obj_type = self.__resolve_object_args(
                                obj_args=[obj_type],
                                graph_args=graph_args,
                                is_root=is_root)[0]
                            obj_args = self.__resolve_object_args(
                                obj_args=line[5:],
                                graph_args=graph_args,
                                is_root=is_root
                            )

                            # replace args with resolved args
                            line = line[:5] + obj_args

                        abs_path = self.find_abstraction_path(pd_path.parent, obj_type)

                        if abs_path is not None:
                            g, gop = self.gui_from_file(abs_path, obj_args=obj_args, is_root=False)

                            if gop:
                                assert isinstance(g, Graph)
                                # set object coordinates
                                g.position = Coords(
                                    x=int(line[2]),
                                    y=int(line[3])
                                )
                                graphs.append(g)

                        if obj_type == "cnv":
                            x = self.add_canvas(line)
                        elif obj_type == "bng":
                            x = self.add_bang(line)
                        elif obj_type == "tgl":
                            x = self.add_toggle(line)
                        elif obj_type == "vradio" or obj_type == "hradio":
                            x = self.add_radio(line)
                        elif obj_type == "vsl" or obj_type == "hsl":
                            x = self.add_slider(line)
                        elif obj_type == "knob" or obj_type == "else/knob":
                            x = self.add_knob(line)
                        elif obj_type == "nbx":
                            x = self.add_number(line)

                    if x is not None:
                        objects.append(x)
                        x = None

        except Exception as e:
            raise e

        if is_root:
            line = self.split_line(canvas_line)

            gop_start = Coords(x=0, y=0)
            gop_size = Size(
                x=int(line[4]),
                y=int(line[5])
            )

            objects = self.filter_invisible_objects(objects, gop_start, gop_size)
            graphs = self.filter_invisible_graphs(graphs, gop_start, gop_size)

            return GraphRoot(
                size=gop_size,
                objects=objects,
                graphs=graphs
            ), gop
        else:
            objects = self.filter_invisible_objects(objects, gop_start, gop_size)
            graphs = self.filter_invisible_graphs(graphs, gop_start, gop_size)
            self.object_counter["graph"] += 1

            return Graph(
                id=f"graph{self.object_counter['graph']}",
                position=Coords(
                    x=int(line[2]),
                    y=int(line[3])
                ),
                gop_start=gop_start,
                gop_size=gop_size,
                objects=objects,
                graphs=graphs
            ), gop

    @classmethod
    def filter_invisible_objects(
        cls,
        objects: list[GUIObjects],
        gop_start: Coords,
        gop_size: Size
    ) -> list[GUIObjects]:
        filtered_objects: list[GUIObjects] = []
        gop_max_y = gop_start.y + gop_size.y
        gop_max_x = gop_start.x + gop_size.x

        for obj in objects:
            obj_min_y = obj.position.y
            obj_max_y = obj.position.y + obj.size.y
            obj_min_x = obj.position.x
            obj_max_x = obj.position.x + obj.size.x

            # check if object is overlapping with gop
            if ((obj_min_y < gop_max_y) or (obj_min_x < gop_max_x)) and \
                    ((obj_max_y > gop_start.y) and (obj_max_x > gop_start.x)) and \
                    ((obj_min_y < gop_max_y) and (obj_min_x < gop_max_x)):
                filtered_objects.append(obj)

        return filtered_objects

    @classmethod
    def filter_invisible_graphs(
        cls,
        graphs: list[Graph],
        gop_start: Coords,
        gop_size: Size
    ) -> list[Graph]:
        filtered_graphs: list[Graph] = []
        gop_max_y = gop_start.y + gop_size.y
        gop_max_x = gop_start.x + gop_size.x

        for g in graphs:
            # skip empty graphs
            if cls.graph_is_empty(g):
                continue

            g_min_y = g.position.y
            g_max_y = g.position.y + g.gop_size.y
            g_min_x = g.position.x
            g_max_x = g.position.x + g.gop_size.x

            # check if graph is overlapping with gop
            if ((g_min_y < gop_max_y) or (g_min_x < gop_max_x)) and \
                    ((g_max_y > gop_start.y) and (g_max_x > gop_start.x)) and \
                    ((g_min_y < gop_max_y) and (g_min_x < gop_max_x)):
                filtered_graphs.append(g)

        return filtered_graphs

    @classmethod
    def graph_is_empty(cls, graph: Graph) -> bool:
        """ A graph is considered empty when it doesn't have any objects
            and all of it's child graphs are also empty.
        """
        g_empty: list[bool] = []

        for g in graph.graphs:
            g_empty.append(cls.graph_is_empty(g))

        return len(graph.objects) == 0 and (
            (len(g_empty) and all(g_empty)) or len(graph.graphs) == 0
        )

    @classmethod
    def filter_params(cls, param: str) -> Optional[str]:
        """ Only allow externed parameters
        """

        if "@hv_param" in param or "@hv_event" in param:
            return param.split(" ")[0]
        else:
            return None

    @classmethod
    def __resolve_object_args(
        cls,
        obj_args: list,
        graph_args: list,
        raise_on_failure: bool = True,
        is_root: bool = False
    ) -> list:
        """ Resolve object arguments against the given graph arguments.
            By default this function raises an Exception if it cannot resolve an
            argument.
            This behaviour may be disabled, in which case the unresolved argument
            is replaced with None (which is an otherwise invalid value). A value of
            None typically indicates to a HeavyObject that the default value
            may be used.
        """
        resolved_obj_args = list(obj_args)  # make a copy of the original obj_args
        for i, a in enumerate(obj_args):
            for m in set(cls.RE_DOLLAR.findall(a)):
                assert isinstance(a, str)
                x = int(m)  # the dollar index (i.e. $x)
                if len(graph_args) > x:
                    a = a.replace(fr"\${m}", str(graph_args[x]))

                elif is_root:
                    # NOTE(mhroth): this case is questionable, but since Pd
                    # defaults to this behavior without warning, so will we.
                    a = a.replace(fr"\${m}", "0")

                else:
                    if raise_on_failure:
                        # NOTE(mhroth): this case is questionable, but since Pd
                        # defaults to this behavior without warning, so will we.
                        a = a.replace(fr"\${m}", "0")
                    else:
                        a = None  # indicate that this argument could not be resolved by replacing it with None
            resolved_obj_args[i] = a
        return resolved_obj_args

    def add_canvas(self, line: list[str]) -> Canvas:
        self.object_counter["canvas"] += 1

        label = Label(
            text=line[10],
            color=Color(line[16]),
            position=Coords(
                x=int(line[11]),
                y=int(line[12])
            ),
            font=Font(int(line[13])),
            font_size=int(line[14]) + 2
        ) if line[10] != "empty" else None

        return Canvas(
            id=f"canvas{self.object_counter['canvas']}",
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            label=label,
            size=Size(
                x=int(line[6]),
                y=int(line[7])
            ),
            bg_color=Color(line[15])
        )

    @classmethod
    def add_bang(cls, line: list[str]) -> Optional[Bang]:
        param = cls.filter_params(line[10])
        if param is None:
            return None

        label = Label(
            text=line[11],
            color=Color(line[18]),
            position=Coords(
                x=int(line[12]),
                y=int(line[13])
            ),
            font=Font(int(line[14])),
            font_size=int(line[15]) + 2
        ) if line[11] != "empty" else None

        return Bang(
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5]),
                y=int(line[5])
            ),
            parameter=param,
            label=label,
            flash_time=int(line[6]),
            bg_color=Color(line[16]),
            fg_color=Color(line[17])
        )

    @classmethod
    def add_toggle(cls, line: list[str]) -> Optional[Toggle]:
        param = cls.filter_params(line[8])
        if param is None:
            return None

        label = Label(
            text=line[9],
            color=Color(line[16]),
            position=Coords(
                x=int(line[10]),
                y=int(line[11])
            ),
            font=Font(int(line[12])),
            font_size=int(line[13]) + 2
        ) if line[9] != "empty" else None

        return Toggle(
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5]),
                y=int(line[5])
            ),
            parameter=param,
            label=label,
            bg_color=Color(line[14]),
            fg_color=Color(line[15]),
            non_zero=float(line[18])
        )

    @classmethod
    def add_radio(cls, line: list[str]) -> Optional[Union[VRadio, HRadio]]:
        param = cls.filter_params(line[10])
        if param is None:
            return None

        radio_type = line[4]

        radio_obj: dict[str, type[Union[VRadio, HRadio]]] = {
            "vradio": VRadio,
            "hradio": HRadio
        }

        label = Label(
            text=line[11],
            color=Color(line[18]),
            position=Coords(
                x=int(line[12]),
                y=int(line[13])
            ),
            font=Font(int(line[14])),
            font_size=int(line[15]) + 2
        ) if line[11] != "empty" else None

        return radio_obj[radio_type](
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5]) * (int(line[8]) if radio_type == "hradio" else 1),
                y=int(line[5]) * (int(line[8]) if radio_type == "vradio" else 1)
            ),
            parameter=param,
            label=label,
            bg_color=Color(line[16]),
            fg_color=Color(line[17]),
            options=int(line[8])
        )

    @classmethod
    def add_slider(cls, line: list[str]) -> Optional[Union[VSlider, HSlider]]:
        param = cls.filter_params(line[12])
        if param is None:
            return None

        slider: dict[str, type[Union[VSlider, HSlider]]] = {
            "vsl": VSlider,
            "hsl": HSlider
        }

        label = Label(
            text=line[13],
            color=Color(line[20]),
            position=Coords(
                x=int(line[14]),
                y=int(line[15])
            ),
            font=Font(int(line[16])),
            font_size=int(line[17]) + 2
        ) if line[13] != "empty" else None

        return slider[line[4]](
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5]),
                y=int(line[6])
            ),
            parameter=param,
            label=label,
            bg_color=Color(line[18]),
            fg_color=Color(line[19]),
            min=float(line[7]),
            max=float(line[8]),
            logarithmic=bool(int(line[9])),
            steady=bool(int(line[22]))
        )

    @classmethod
    def add_knob(cls, line: list[str]) -> Optional[Knob]:
        param = cls.filter_params(line[11])
        if param is None:
            return None

        log_val = float(line[8])

        if log_val == 0.0:
            log_mode = "lin"
        elif log_val == 1.0:
            log_mode = "log"
        else:
            log_mode = "exp"

        return Knob(
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5]),
                y=int(line[5])
            ),
            parameter=param,
            label_size=int(line[27]),
            label_pos=Coords(
                x=int(line[28]),
                y=int(line[29])
            ),
            label_show=LabelShow(int(line[26])),
            min=float(line[6]),
            max=float(line[7]),
            bg_color=Color(line[12]),
            fg_color=Color(line[14]),
            init_val=float(line[9]),
            ang_range=int(line[20]),
            ang_offset=int(line[21]),
            log_mode=log_mode,
            exp_fact=float(line[8]),
            discrete=bool(int(line[18])),
            ticks=bool(int(line[32])),
            steps=int(line[17]),
            circular=bool(int(line[16])),
            jump=bool(0),
            square=bool(int(line[15])),
            arc_color=Color(line[13]),
            arc_start=float(line[23]),
            arc_show=bool(int(line[19]))
        )

    @classmethod
    def add_number(cls, line: list[str]) -> Optional[Number]:
        param = cls.filter_params(line[12])
        if param is None:
            return None

        label = Label(
            text=line[13],
            color=Color(line[20]),
            position=Coords(
                x=int(line[14]),
                y=int(line[15])
            ),
            font=Font(0),
            font_size=int(line[17]) + 2
        ) if line[13] != "empty" else None

        return Number(
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[5])*int(line[17])+8,
                y=int(line[6])
            ),
            parameter=param,
            label=label,
            bg_color=Color(line[18]),
            fg_color=Color(line[19]),
            log_mode=bool(int(line[9])),
            log_height=int(line[21]),
            min=float(line[7]),
            max=float(line[8])
        )

    def add_comment(self, line: list[str]) -> Comment:
        self.object_counter["comment"] += 1

        # get width and clean up remaining text
        if line[-2] == "f" \
                and line[-3][-1] == ",":
            width = int(line[-1])
            text = " ".join(line[4:-2])[:-1]
        else:
            width = None
            text = " ".join(line[4:])

        # escape characters
        text = text.replace("\"", "\\\"")
        # text = text.replace("\\": "\\\\")  # backslash is already removed by self.split_line()

        return Comment(
            id=f"comment{self.object_counter['comment']}",
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            text=text,
            width=width,
            size=Size(
                x=10*len(text), y=10
            )
        )

    @classmethod
    def add_float(cls, line: list[str]) -> Optional[Float]:
        param = cls.filter_params(line[9])
        if param is None:
            return None

        font_height = int(line[11]) if line[11] != "0" else 8

        return Float(
            position=Coords(
                x=int(line[2]),
                y=int(line[3])
            ),
            size=Size(
                x=int(line[4]) * (font_height - 4),
                y=font_height
            ),
            parameter=param,
            label_text=line[8] if line[8] != "-" else "",
            font_height=font_height,
            label_pos=LabelPos(int(line[7])),
            min=float(line[5]),
            max=float(line[6])
        )



# Heavy Compiler Collection
# Copyright (C) 2026 Wasted Audio
#
# SPDX-License-Identifier: GPL-3.0-only

from typing import Union

from pydantic import BaseModel, Field, ConfigDict


class HvPos(BaseModel):
    x: int
    y: int


class HvConnFrom(BaseModel):
    id: str
    outlet: int


class HvConnTo(BaseModel):
    id: str
    inlet: int


class HvConn(BaseModel):
    model_config = ConfigDict(populate_by_name=True)

    type: str
    conn_from: HvConnFrom = Field(..., alias='from')
    conn_to: HvConnTo = Field(..., alias='to')


class Heavy(BaseModel):
    type: str
    imports: list = Field(default_factory=list)
    args: Union[list, dict] = Field(default_factory=list)  # we should make this more specific
    objects: dict[str, "Heavy"] = Field(default_factory=dict)
    connections: list[HvConn] = Field(default_factory=list)
    properties: HvPos = Field(default_factory=lambda: HvPos(x=0, y=0))
    annotations: dict = Field(default_factory=dict)


Heavy.model_rebuild()



# Copyright (C) 2026 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional
from pathlib import Path

from .HeavyObject import HeavyObject
from .SignalNamHeader import convert_nam_to_header, to_symbol_stem
from hvcc.types.IR import IRSignalList


class UnknownModelException(Exception):
    def __str__(self):
        return "Unknown NAM model used."


class SignalNam(HeavyObject):

    preamble = "sNam"

    @classmethod
    def get_C_struct(cls, obj_type: str = "") -> str:
        if obj_type == '__nam_nano~f':
            return "SignalNamNano"
        elif obj_type == '__nam_feather~f':
            return "SignalNamFeather"
        elif obj_type == '__nam_lite~f':
            return "SignalNamLite"
        elif obj_type == '__nam_standard~f':
            return "SignalNamStandard"
        else:
            raise UnknownModelException()

    @classmethod
    def get_C_header_set(cls) -> set:
        return {"HvSignalNam.h"}

    @classmethod
    def get_C_file_set(cls) -> set:
        return {
            "HvSignalNam.h",
            "HvSignalNam.c",
            "MicroNAM_C.h",
            "MicroNAM_C.cpp",
            "MicroNam/MicroNAM.h",
            "MicroNam/include/StandardNet.h",
            "MicroNam/include/NanoNet.h",
            "MicroNam/include/LiteNet.h",
            "MicroNam/include/FeatherNet.h"
        }

    @classmethod
    def get_C_init(cls, obj_type: str, obj_id: str, args: dict) -> list[str]:
        nam_file = Path(args["nam"])
        weights = to_symbol_stem(nam_file)

        if obj_type == '__nam_nano~f':
            return [f"sNam_nano_init(&sNam_{obj_id}, {weights}Weights);"]
        elif obj_type == '__nam_feather~f':
            return [f"sNam_feather_init(&sNam_{obj_id}, {weights}Weights);"]
        elif obj_type == '__nam_lite~f':
            return [f"sNam_lite_init(&sNam_{obj_id}, {weights}Weights);"]
        elif obj_type == '__nam_standard~f':
            return [f"sNam_standard_init(&sNam_{obj_id}, {weights}Weights);"]
        else:
            raise UnknownModelException()

    @classmethod
    def get_C_free(cls, obj_type: str, obj_id: str, args: dict) -> list[str]:
        if obj_type == '__nam_nano~f':
            return [f"sNam_nano_free(&sNam_{obj_id});"]
        elif obj_type == '__nam_feather~f':
            return [f"sNam_feather_free(&sNam_{obj_id});"]
        elif obj_type == '__nam_lite~f':
            return [f"sNam_lite_free(&sNam_{obj_id});"]
        elif obj_type == '__nam_standard~f':
            return [f"sNam_standard_free(&sNam_{obj_id});"]
        else:
            raise UnknownModelException()

    @classmethod
    def get_C_process(cls, process_dict: IRSignalList, obj_type: str, obj_id: str, args: dict) -> list[str]:
        if obj_type == '__nam_nano~f':
            function = "__hv_nam_nano_f(&sNam_{0}, VIf({1}), VOf({2}));"
        elif obj_type == '__nam_feather~f':
            function = "__hv_nam_feather_f(&sNam_{0}, VIf({1}), VOf({2}));"
        elif obj_type == '__nam_lite~f':
            function = "__hv_nam_lite_f(&sNam_{0}, VIf({1}), VOf({2}));"
        elif obj_type == '__nam_standard~f':
            function = "__hv_nam_standard_f(&sNam_{0}, VIf({1}), VOf({2}));"
        else:
            raise UnknownModelException()

        return [
            function.format(
                process_dict.id,
                cls._c_buffer(process_dict.inputBuffers[0]),
                cls._c_buffer(process_dict.outputBuffers[0]))
        ]

    @classmethod
    def get_C_gen_header_code(cls, obj_type: str, obj_id: str, args: dict) -> Optional[tuple[str, str]]:
        nam_file = Path(args["nam"])
        header_name = f"{to_symbol_stem(nam_file)}.h"

        header_str = convert_nam_to_header(nam_file)

        return header_name, header_str



# Copyright (C) 2014-2018 Enzien Audio, Ltd.
# Copyright (C) 2023-2024 Wasted Audio
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Optional, Dict

from .HeavyIrObject import HeavyIrObject
from .HeavyLangObject import HeavyLangObject
from .HeavyGraph import HeavyGraph


class HLangDac(HeavyLangObject):
    """ Translates HeavyLang [dac] to HeavyIR [__add~f].
    """

    def __init__(
        self,
        obj_type: str,
        args: Dict,
        graph: 'HeavyGraph',
        annotations: Optional[Dict] = None
    ) -> None:
        assert obj_type == "dac"
        super().__init__(obj_type, args, graph,
                         num_inlets=len(args[self._HEAVY_LANG_DICT[obj_type].args[0].name]),
                         num_outlets=0,
                         annotations=annotations)

    def reduce(self) -> tuple:
        objects = set()
        connections = []

        # reduce a HeavyLang dac to a number of individual HeavyIR __add~f objects
        for i, channel_index in enumerate(self.args["channels"]):
            if len(self.inlet_connections[i]) > 0:  # if there are any connections to this inlet
                x = HeavyIrObject("__add~f")
                x.inlet_buffers[1] = ("output", channel_index - 1)  # channel indicies are one-indexed
                x.outlet_buffers[0] = ("output", channel_index - 1)
                objects.add(x)

                for c in self.inlet_connections[i]:
                    connections.append((c, [c.copy(to_object=x, inlet_index=0)]))

        return (objects, connections)
